"""Tests for the Chirps module."""
import calendar
from datetime import date
from pathlib import Path

import cftime
import numpy as np
import pytest
import xarray as xr
from xarray.coding.cftimeindex import CFTimeIndex

from aatoolbox import ChirpsDaily, ChirpsMonthly, GeoBoundingBox

DATASOURCE_BASE_DIR = "chirps"

START_DATE = date(year=2020, month=7, day=15)
END_DATE = date(year=2020, month=8, day=1)
CURRENT_DATE = date(year=2022, month=5, day=31)
PAST_DATE = date(year=1970, month=1, day=1)
FUTURE_DATE = date(year=2100, month=10, day=2)

START_YEAR = "2020"
START_MONTH = "07"
START_MONTH_NAME = "Jul"
START_DAY = "15"

END_YEAR = "2020"
END_MONTH = "08"
END_MONTH_NAME = "Aug"
END_DAY = "01"


@pytest.fixture
def mock_chirps(mocker, mock_country_config):
    """Create Chirps class with mock country config."""
    geo_bounding_box = GeoBoundingBox(
        lat_max=6, lat_min=3.2, lon_max=2, lon_min=-3
    )

    mocker.patch(
        (
            "aatoolbox.datasources.chirps.chirps.ChirpsDaily"
            "._get_last_available_date"
        ),
        return_value=CURRENT_DATE,
    )

    mocker.patch(
        (
            "aatoolbox.datasources.chirps.chirps.ChirpsMonthly"
            "._get_last_available_date"
        ),
        return_value=CURRENT_DATE,
    )

    def _mock_chirps(
        frequency: str = "daily",
        resolution: float = 0.05,
        start_date: date = START_DATE,
        end_date: date = END_DATE,
    ):
        if frequency == "daily":
            chirps = ChirpsDaily(
                country_config=mock_country_config,
                geo_bounding_box=geo_bounding_box,
                resolution=resolution,
                start_date=start_date,
                end_date=end_date,
            )
        else:
            chirps = ChirpsMonthly(
                country_config=mock_country_config,
                geo_bounding_box=geo_bounding_box,
                start_date=start_date,
                end_date=end_date,
            )
        return chirps

    return _mock_chirps


@pytest.fixture
def mock_xr_open_multiple_dataset(mocker):
    """Mock GeoPandas file reading function."""
    ds = xr.Dataset()
    return mocker.patch(
        "aatoolbox.datasources.chirps.chirps.xr.open_mfdataset",
        return_value=ds,
    )


@pytest.fixture
def mock_dataset_to_netcdf(mocker):
    """Mock GeoPandas file writing function."""
    return mocker.patch(
        "aatoolbox.datasources.chirps.chirps.xr.Dataset.to_netcdf",
    )


@pytest.fixture
def mock_download(mocker, mock_chirps):
    """Create mock for download method."""
    download_mock = mocker.patch(
        "aatoolbox.datasources.chirps.chirps._Chirps._download"
    )

    def _mock_download(
        frequency: str,
    ):
        chirps = mock_chirps(
            frequency=frequency,
            start_date=START_DATE,
            end_date=END_DATE,
        )
        chirps.download()
        args_download = download_mock.call_args_list
        url_list = [k["url"] for (_, k) in args_download]
        filepath_list = [k["filepath"] for (_, k) in args_download]
        return url_list, filepath_list

    return _mock_download


def test_valid_arguments_class(mocker, mock_country_config):
    """Test for resolution and wrong dates in initialisation class."""
    mocker.patch(
        (
            "aatoolbox.datasources.chirps.chirps.ChirpsDaily"
            "._get_last_available_date"
        ),
        return_value=CURRENT_DATE,
    )

    geo_bounding_box = GeoBoundingBox(
        lat_max=6, lat_min=3.2, lon_max=2, lon_min=-3
    )
    with pytest.raises(ValueError):
        ChirpsDaily(
            country_config=mock_country_config,
            geo_bounding_box=geo_bounding_box,
            resolution=0.10,
            start_date=START_DATE,
            end_date=END_DATE,
        )
        ChirpsDaily(
            country_config=mock_country_config,
            geo_bounding_box=geo_bounding_box,
            resolution=0.10,
            start_date=START_DATE,
            end_date=FUTURE_DATE,
        )
        ChirpsDaily(
            country_config=mock_country_config,
            geo_bounding_box=geo_bounding_box,
            resolution=0.10,
            start_date=PAST_DATE,
            end_date=END_DATE,
        )


def test_download_monthly(
    mock_aa_data_dir, mock_country_config, mock_download
):
    """Test of call download for monthly data."""
    url_list, filepath_list = mock_download(frequency="monthly")

    url_list_control = [
        (
            "https://iridl.ldeo.columbia.edu/SOURCES/.UCSB/"
            ".CHIRPS/.v2p0/.monthly/.global/.precipitation/"
            f"X/%28-3.0%29%282.0%29RANGEEDGES/"
            f"Y/%286.0%29%283.2%29RANGEEDGES/"
            f"T/%28{MONTH_NAME}%20{START_YEAR}%29"
            f"%28{MONTH_NAME}%20{START_YEAR}"
            "%29RANGEEDGES/data.nc"
        )
        for MONTH_NAME in [
            calendar.month_abbr[month]
            for month in range(int(START_MONTH), int(END_MONTH) + 1)
        ]
    ]

    filepath_list_control = [
        mock_aa_data_dir
        / (
            f"public/raw/{mock_country_config.iso3}/{DATASOURCE_BASE_DIR}/"
            f"abc_chirps_monthly_{START_YEAR}_{MONTH:02d}_"
            "r0.05_Np6Sp3Ep2Wm3.nc"
        )
        for MONTH in range(int(START_MONTH), int(END_MONTH) + 1)
    ]

    assert url_list == url_list_control
    assert filepath_list == filepath_list_control


def test_download_daily(mock_aa_data_dir, mock_country_config, mock_download):
    """Test of call download for daily data."""
    url_list, filepath_list = mock_download(frequency="daily")

    url_list_control = [
        (
            "https://iridl.ldeo.columbia.edu/SOURCES/.UCSB/"
            ".CHIRPS/.v2p0/.daily-improved/.global/.0p05/.prcp/"
            f"X/%28-3.0%29%282.0%29RANGEEDGES/"
            f"Y/%286.0%29%283.2%29RANGEEDGES/"
            f"T/%28{DAY:02d}%20{START_MONTH_NAME}%20{START_YEAR}"
            f"%29%28{DAY:02d}%20{START_MONTH_NAME}%20{START_YEAR}"
            "%29RANGEEDGES/data.nc"
        )
        for DAY in range(int(START_DAY), 32)
    ]

    url_list_control.append(
        "https://iridl.ldeo.columbia.edu/SOURCES/.UCSB/"
        ".CHIRPS/.v2p0/.daily-improved/.global/.0p05/.prcp/"
        f"X/%28-3.0%29%282.0%29RANGEEDGES/"
        f"Y/%286.0%29%283.2%29RANGEEDGES/"
        f"T/%2801%20{END_MONTH_NAME}%20{START_YEAR}"
        f"%29%2801%20{END_MONTH_NAME}%20{START_YEAR}"
        "%29RANGEEDGES/data.nc"
    )

    filepath_list_control = [
        mock_aa_data_dir
        / (
            f"public/raw/{mock_country_config.iso3}/{DATASOURCE_BASE_DIR}/"
            f"abc_chirps_daily_{START_YEAR}_{START_MONTH}_{DAY}_"
            "r0.05_Np6Sp3Ep2Wm3.nc"
        )
        for DAY in range(int(START_DAY), 32)
    ]

    filepath_list_control.append(
        mock_aa_data_dir
        / (
            f"public/raw/{mock_country_config.iso3}/{DATASOURCE_BASE_DIR}/"
            f"abc_chirps_daily_{START_YEAR}_{END_MONTH}_{END_DAY}_"
            "r0.05_Np6Sp3Ep2Wm3.nc"
        )
    )

    assert url_list == url_list_control
    assert filepath_list == filepath_list_control


def test_process_monthly(
    mocker,
    mock_chirps,
    mock_aa_data_dir,
    mock_country_config,
    mock_dataset_to_netcdf,
):
    """Test process monthl data."""
    ds = xr.DataArray(
        np.reshape(a=np.arange(8), newshape=(2, 2, 2)),
        dims=("X", "Y", "T"),
        coords={
            "X": [1, 2],
            "Y": [2, -3],
            "T": [685.5, 686.5],
        },
    ).to_dataset(name="prcp")

    ds["T"].attrs["calendar"] = "360"
    ds["T"].attrs["units"] = "months since 1960-01-01"

    mock_xr_open_dataset = mocker.patch(
        "aatoolbox.datasources.chirps.chirps.xr.open_dataset",
        return_value=ds,
    )

    input_filepath_list_control = [
        mock_aa_data_dir
        / (
            f"public/raw/{mock_country_config.iso3}/"
            f"{DATASOURCE_BASE_DIR}/"
            f"abc_chirps_monthly_{START_YEAR}_{MONTH:02d}_"
            "r0.05_Np6Sp3Ep2Wm3.nc"
        )
        for MONTH in range(int(START_MONTH), int(END_MONTH) + 1)
    ]

    output_filepath_list_control = [
        Path(str(f).replace("raw", "processed"))
        for f in input_filepath_list_control
    ]

    chirps = mock_chirps(frequency="monthly")
    processed_path = chirps.process()

    args_input_process = mock_xr_open_dataset.call_args_list
    input_filepath_list = [k[0] for (k, _) in args_input_process]

    args_output_process = mock_dataset_to_netcdf.call_args_list
    output_filepath_list = [k["path"] for (_, k) in args_output_process]
    output_dataset_list = [k[0] for (k, _) in args_output_process]
    output_ds = output_dataset_list[0]

    assert input_filepath_list == input_filepath_list_control
    assert output_filepath_list == output_filepath_list_control

    assert processed_path == (
        mock_aa_data_dir
        / f"public/processed/{mock_country_config.iso3}/{DATASOURCE_BASE_DIR}/"
    )

    expected_f = CFTimeIndex(
        [
            cftime.Datetime360Day(year=2017, month=2, day=16),
            cftime.Datetime360Day(year=2017, month=3, day=16),
        ]
    )

    assert np.array_equal(output_ds.X.values, [1, 2])
    assert np.array_equal(output_ds.Y.values, [2, -3])
    assert output_ds.get_index("T").equals(expected_f)
    assert np.array_equal(
        output_ds.precipitation.values,
        np.reshape(a=np.arange(8), newshape=(2, 2, 2)),
    )


def test_process_daily(
    mocker,
    mock_chirps,
    mock_aa_data_dir,
    mock_country_config,
    mock_dataset_to_netcdf,
):
    """Test process daily data."""
    ds = xr.DataArray(
        np.reshape(a=np.arange(8), newshape=(2, 2, 2)),
        dims=("X", "Y", "T"),
        coords={
            "X": [1, 2],
            "Y": [2, -3],
            "T": [2454397.0, 2454398.0],
        },
    ).to_dataset(name="prcp")

    ds["T"].attrs["calendar"] = "standard"
    ds["T"].attrs["units"] = "julian_day"

    mock_xr_open_dataset = mocker.patch(
        "aatoolbox.datasources.chirps.chirps.xr.open_dataset",
        return_value=ds,
    )

    input_filepath_list_control = [
        mock_aa_data_dir
        / (
            f"public/raw/{mock_country_config.iso3}/"
            f"{DATASOURCE_BASE_DIR}/"
            f"abc_chirps_daily_{START_YEAR}_{START_MONTH}_{DAY}_"
            "r0.05_Np6Sp3Ep2Wm3.nc"
        )
        for DAY in range(int(START_DAY), 32)
    ]

    input_filepath_list_control.append(
        mock_aa_data_dir
        / (
            f"public/raw/{mock_country_config.iso3}/"
            f"{DATASOURCE_BASE_DIR}/"
            f"abc_chirps_daily_{START_YEAR}_{END_MONTH}_{END_DAY}_"
            "r0.05_Np6Sp3Ep2Wm3.nc"
        )
    )

    output_filepath_list_control = [
        Path(str(f).replace("raw", "processed"))
        for f in input_filepath_list_control
    ]

    chirps = mock_chirps(frequency="daily")
    processed_path = chirps.process()

    args_input_process = mock_xr_open_dataset.call_args_list
    input_filepath_list = [k[0] for (k, _) in args_input_process]

    args_output_process = mock_dataset_to_netcdf.call_args_list
    output_filepath_list = [k["path"] for (_, k) in args_output_process]
    output_dataset_list = [k[0] for (k, _) in args_output_process]
    output_ds = output_dataset_list[0]

    assert input_filepath_list == input_filepath_list_control
    assert output_filepath_list == output_filepath_list_control

    assert processed_path == (
        mock_aa_data_dir
        / f"public/processed/{mock_country_config.iso3}/{DATASOURCE_BASE_DIR}/"
    )

    expected_f = CFTimeIndex(
        [
            cftime.DatetimeGregorian(year=2007, month=10, day=23, hour=12),
            cftime.DatetimeGregorian(year=2007, month=10, day=24, hour=12),
        ]
    )

    assert np.array_equal(output_ds.X.values, [1, 2])
    assert np.array_equal(output_ds.Y.values, [2, -3])
    assert output_ds.get_index("T").equals(expected_f)
    assert np.array_equal(
        output_ds.precipitation.values,
        np.reshape(a=np.arange(8), newshape=(2, 2, 2)),
    )


def test_chirps_load_monthly(
    mock_xr_open_multiple_dataset,
    mock_chirps,
    mock_aa_data_dir,
    mock_country_config,
):
    """Test load monthly data."""
    filepath_list_control = [
        mock_aa_data_dir
        / (
            f"public/processed/{mock_country_config.iso3}/"
            f"{DATASOURCE_BASE_DIR}/"
            f"abc_chirps_monthly_{START_YEAR}_{MONTH:02d}_"
            "r0.05_Np6Sp3Ep2Wm3.nc"
        )
        for MONTH in range(int(START_MONTH), int(END_MONTH) + 1)
    ]

    chirps = mock_chirps(frequency="monthly")
    chirps.load()
    args_download = mock_xr_open_multiple_dataset.call_args
    filepath_list = args_download[0][0]

    ds = mock_xr_open_multiple_dataset()

    assert ds.attrs["included_files"] == [
        filepath.stem for filepath in filepath_list
    ]
    assert filepath_list == filepath_list_control


def test_chirps_load_daily(
    mock_xr_open_multiple_dataset,
    mock_chirps,
    mock_aa_data_dir,
    mock_country_config,
):
    """Test load daily data."""
    filepath_list_control = [
        mock_aa_data_dir
        / (
            f"public/processed/{mock_country_config.iso3}/"
            f"{DATASOURCE_BASE_DIR}/"
            f"abc_chirps_daily_{START_YEAR}_{START_MONTH}_{DAY}_"
            "r0.05_Np6Sp3Ep2Wm3.nc"
        )
        for DAY in range(int(START_DAY), 32)
    ]

    filepath_list_control.append(
        mock_aa_data_dir
        / (
            f"public/processed/{mock_country_config.iso3}/"
            f"{DATASOURCE_BASE_DIR}/"
            f"abc_chirps_daily_{START_YEAR}_{END_MONTH}_{END_DAY}_"
            "r0.05_Np6Sp3Ep2Wm3.nc"
        )
    )

    chirps = mock_chirps(frequency="daily")
    chirps.load()
    args_download = mock_xr_open_multiple_dataset.call_args
    filepath_list = args_download[0][0]

    ds = mock_xr_open_multiple_dataset()

    assert ds.attrs["included_files"] == [
        filepath.stem for filepath in filepath_list
    ]
    assert filepath_list == filepath_list_control



"""Classes to download and process USGS FEWS NET NDVI data.

Download, process, and load NDVI data published
in the `USGS FEWS NET data portal
<https://earlywarning.usgs.gov/fews>`_. Classes available
to process temporally smoothed NDVI values, percent of
median values, difference to median values, and difference
from current value to the previous year's value.
"""

# TODO: add progress bar
import logging
from datetime import date
from typing import Tuple, Union

from aatoolbox.config.countryconfig import CountryConfig
from aatoolbox.datasources.usgs.ndvi_base import _UsgsNdvi

logger = logging.getLogger(__name__)


class UsgsNdviSmoothed(_UsgsNdvi):
    """Base class to retrieve smoothed NDVI data.

    The retrieved data is the smoothed NDVI values
    processed by the USGS. Temporal smoothing is done
    to adjust for cloud cover and other errors.
    Data for the 3 most recent dekads is not fully
    smoothed, and are re-smoothed at the end of the
    3 dekad period.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    start_date : _DATE_TYPE, default = None
        Start date. Can be passed as a ``datetime.date``
        object or a data string in ISO8601 format, and
        the relevant dekad will be determined. Or pass
        directly as year-dekad tuple, e.g. (2020, 1).
        If ``None``, ``start_date`` is set to earliest
        date with data: 2002, dekad 19.
    end_date : _DATE_TYPE, default = None
        End date. Can be passed as a ``datetime.date``
        object and the relevant dekad will be determined,
        as a date string in ISO8601 format, or as a
        year-dekad tuple, i.e. (2020, 1). If ``None``,
        ``end_date`` is set to ``date.today()``.

    Examples
    --------
    >>> from aatoolbox import create_country_config, \
    ...  CodAB, UsgsNdviSmoothed
    >>>
    >>> # Retrieve admin 2 boundaries for Burkina Faso
    >>> country_config = create_country_config(iso3="bfa")
    >>> codab = CodAB(country_config=country_config)
    >>> bfa_admin2 = codab.load(admin_level=2)
    >>>
    >>> # setup NDVI
    >>> bfa_ndvi = UsgsNdviSmoothed(
    ...     country_config=country_config,
    ...     start_date=[2020, 1],
    ...     end_date=[2020, 3]
    ... )
    >>> bfa_ndvi.download()
    >>> bfa_ndvi.process(
    ...     gdf=bfa_admin2,
    ...     feature_col="ADM2_FR"
    ... )
    >>>
    >>> # load in processed data
    >>> df = bfa_ndvi.load(feature_col="ADM2_FR")
    """

    def __init__(
        self,
        country_config: CountryConfig,
        start_date: Union[date, str, Tuple[int, int], None] = None,
        end_date: Union[date, str, Tuple[int, int], None] = None,
    ):
        super().__init__(
            country_config=country_config,
            data_variable="smoothed",
            data_variable_suffix="",
            data_variable_url="temporallysmoothedndvi",
            start_date=start_date,
            end_date=end_date,
        )


class UsgsNdviPctMedian(_UsgsNdvi):
    """Base class to retrieve % of median NDVI.

    The retrieved data is the percent of median NDVI
    values calculated from 2003 - 2017, as
    processed by the USGS.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    start_date : _DATE_TYPE, default = None
        Start date. Can be passed as a ``datetime.date``
        object or a data string in ISO8601 format, and
        the relevant dekad will be determined. Or pass
        directly as year-dekad tuple, e.g. (2020, 1).
        If ``None``, ``start_date`` is set to earliest
        date with data: 2002, dekad 19.
    end_date : _DATE_TYPE, default = None
        End date. Can be passed as a ``datetime.date``
        object and the relevant dekad will be determined,
        as a date string in ISO8601 format, or as a
        year-dekad tuple, i.e. (2020, 1). If ``None``,
        ``end_date`` is set to ``date.today()``.

    Examples
    --------
    >>> from aatoolbox import create_country_config, \
    ...  CodAB, UsgsNdviPctMedian
    >>>
    >>> # Retrieve admin 2 boundaries for Burkina Faso
    >>> country_config = create_country_config(iso3="bfa")
    >>> codab = CodAB(country_config=country_config)
    >>> bfa_admin2 = codab.load(admin_level=2)
    >>>
    >>> # setup NDVI
    >>> bfa_ndvi = UsgsNdviPctMedian(
    ...     country_config=country_config,
    ...     start_date=[2020, 1],
    ...     end_date=[2020, 3]
    ... )
    >>> bfa_ndvi.download()
    >>> bfa_ndvi.process(
    ...     gdf=bfa_admin2,
    ...     feature_col="ADM2_FR"
    ... )
    >>>
    >>> # load in processed data
    >>> df = bfa_ndvi.load(feature_col="ADM2_FR")
    """

    def __init__(
        self,
        country_config: CountryConfig,
        start_date: Union[date, str, Tuple[int, int], None] = None,
        end_date: Union[date, str, Tuple[int, int], None] = None,
    ):
        super().__init__(
            country_config=country_config,
            data_variable="percent_median",
            data_variable_suffix="pct",
            data_variable_url="percentofmedian",
            start_date=start_date,
            end_date=end_date,
        )


class UsgsNdviMedianAnomaly(_UsgsNdvi):
    """Base class to retrieve NDVI anomaly data.

    The retrieved data is NDVI anomaly data calculated
    as a subtraction of the median value, based on data
    from 2003 - 2017, from the
    current value. Negative values indicate less
    vegetation than the median, positive values indicate
    more vegetation.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    start_date : _DATE_TYPE, default = None
        Start date. Can be passed as a ``datetime.date``
        object or a data string in ISO8601 format, and
        the relevant dekad will be determined. Or pass
        directly as year-dekad tuple, e.g. (2020, 1).
        If ``None``, ``start_date`` is set to earliest
        date with data: 2002, dekad 19.
    end_date : _DATE_TYPE, default = None
        End date. Can be passed as a ``datetime.date``
        object and the relevant dekad will be determined,
        as a date string in ISO8601 format, or as a
        year-dekad tuple, i.e. (2020, 1). If ``None``,
        ``end_date`` is set to ``date.today()``.

    Examples
    --------
    >>> from aatoolbox import create_country_config, \
    ...  CodAB, UsgsNdviMedianAnomaly
    >>>
    >>> # Retrieve admin 2 boundaries for Burkina Faso
    >>> country_config = create_country_config(iso3="bfa")
    >>> codab = CodAB(country_config=country_config)
    >>> bfa_admin2 = codab.load(admin_level=2)
    >>>
    >>> # setup NDVI
    >>> bfa_ndvi = UsgsNdviMedianAnomaly(
    ...     country_config=country_config,
    ...     start_date=[2020, 1],
    ...     end_date=[2020, 3]
    ... )
    >>> bfa_ndvi.download()
    >>> bfa_ndvi.process(
    ...     gdf=bfa_admin2,
    ...     feature_col="ADM2_FR"
    ... )
    >>>
    >>> # load in processed data
    >>> df = bfa_ndvi.load(feature_col="ADM2_FR")
    """

    def __init__(
        self,
        country_config: CountryConfig,
        start_date: Union[date, str, Tuple[int, int], None] = None,
        end_date: Union[date, str, Tuple[int, int], None] = None,
    ):
        super().__init__(
            country_config=country_config,
            data_variable="median_anomaly",
            data_variable_suffix="stmdn",
            data_variable_url="mediananomaly",
            start_date=start_date,
            end_date=end_date,
        )


class UsgsNdviYearDifference(_UsgsNdvi):
    """Base class to retrieve NDVI year difference data.

    The retrieved data is NDVI yearly difference data,
    calculated as the subtraction of the previous year's
    NDVI value from the current year's. Negative
    values indicate the current vegetation is less
    than the previous year's, positive that there
    is more vegetation in the current year.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    start_date : _DATE_TYPE, default = None
        Start date. Can be passed as a ``datetime.date``
        object or a data string in ISO8601 format, and
        the relevant dekad will be determined. Or pass
        directly as year-dekad tuple, e.g. (2020, 1).
        If ``None``, ``start_date`` is set to earliest
        date with data: 2002, dekad 19.
    end_date : _DATE_TYPE, default = None
        End date. Can be passed as a ``datetime.date``
        object and the relevant dekad will be determined,
        as a date string in ISO8601 format, or as a
        year-dekad tuple, i.e. (2020, 1). If ``None``,
        ``end_date`` is set to ``date.today()``.

    Examples
    --------
    >>> from aatoolbox import create_country_config, \
    ...  CodAB, UsgsNdviDifference
    >>>
    >>> # Retrieve admin 2 boundaries for Burkina Faso
    >>> country_config = create_country_config(iso3="bfa")
    >>> codab = CodAB(country_config=country_config)
    >>> bfa_admin2 = codab.load(admin_level=2)
    >>>
    >>> # setup NDVI
    >>> bfa_ndvi = UsgsNdviDifference(
    ...     country_config=country_config,
    ...     start_date=[2020, 1],
    ...     end_date=[2020, 3]
    ... )
    >>> bfa_ndvi.download()
    >>> bfa_ndvi.process(
    ...     gdf=bfa_admin2,
    ...     feature_col="ADM2_FR"
    ... )
    >>>
    >>> # load in processed data
    >>> df = bfa_ndvi.load(feature_col="ADM2_FR")
    """

    def __init__(
        self,
        country_config: CountryConfig,
        start_date: Union[date, str, Tuple[int, int], None] = None,
        end_date: Union[date, str, Tuple[int, int], None] = None,
    ):
        super().__init__(
            country_config=country_config,
            data_variable="difference",
            data_variable_suffix="dif",
            data_variable_url="differencepreviousyear",
            start_date=start_date,
            end_date=end_date,
        )



"""Tests for GloFAS processing."""
from datetime import date
from typing import List, Union

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from aatoolbox import (
    CountryConfig,
    GeoBoundingBox,
    GlofasForecast,
    GlofasReanalysis,
    GlofasReforecast,
)


class TestProcess:
    """Tests for GloFAS processing."""

    geo_bounding_box = GeoBoundingBox(
        lat_max=1, lat_min=-2, lon_max=3, lon_min=-4
    )
    numbers = [0, 1, 2, 3, 4, 5, 6]

    def get_raw_data(
        self,
        number_coord: Union[List[int], int] = None,
        include_step: bool = False,
        include_history: bool = False,
        dis24: np.ndarray = None,
        single_day=False,
    ) -> xr.Dataset:
        """
        Construct a simple fake GloFAS xarray dataset.

        Parameters
        ----------
        number_coord : list or int, default = None
            The ensemble number coordinate
        include_step :  bool, default = False
            Whether to include the forecast step coordinate
        include_history : bool, default = False
            Whether to include the history attribute
        dis24 : np.ndarray, default = None
            Optional array of discharge values, that should have the combined
            dimensions of the coordinates. If not passed, generated using
            random numbers.
        single_day: boolean, default = False
            Only create a single day of data, required for the forecasts

        Returns
        -------
        Simplified GloFAS xarray dataset
        """
        rng = np.random.default_rng(12345)
        coords = {}
        if number_coord is not None:
            coords["number"] = number_coord
        if single_day:
            coords["time"] = np.datetime64("2014-09-06")
        else:
            coords["time"] = pd.date_range("2014-09-06", periods=2)
        if include_step:
            coords["step"] = [np.datetime64(n + 1, "D") for n in range(5)]
        coords["latitude"] = (
            np.arange(
                start=self.geo_bounding_box.lat_min,
                stop=self.geo_bounding_box.lat_max + 2,
                step=0.1,
            )
            - 0.05
        )
        coords["longitude"] = (
            np.arange(
                start=self.geo_bounding_box.lon_min,
                stop=self.geo_bounding_box.lon_max + 2,
                step=0.1,
            )
            - 0.05
        )
        dims = list(coords.keys())
        if number_coord is not None and isinstance(number_coord, int):
            dims.remove("number")
        if single_day:
            dims.remove("time")
        if dis24 is None:
            dis24 = 5000 + 100 * rng.random([len(coords[dim]) for dim in dims])
        attrs = {}
        if include_history:
            attrs = {"history": "fake history"}
        return xr.Dataset({"dis24": (dims, dis24)}, coords=coords, attrs=attrs)

    @pytest.fixture()
    def mock_ensemble_raw(self):
        """
        Create fake raw ensemble data.

        For the forecast and reforecast, generate the raw data, which consists
        of the control and perturbed forecast, and combine the discharge
        values for the processed data. Return both xarray datasets and the
        combined array.
        """

        def _mock_ensemble_raw(
            single_day: bool = False,
        ) -> (xr.Dataset, xr.Dataset, np.ndarray):
            cf_raw = self.get_raw_data(
                number_coord=self.numbers[0],
                include_step=True,
                include_history=True,
                single_day=single_day,
            )
            pf_raw = self.get_raw_data(
                number_coord=self.numbers[1:],
                include_step=True,
                include_history=True,
                single_day=single_day,
            )
            expected_dis24 = np.concatenate(
                (
                    cf_raw["dis24"].values[np.newaxis, ...],
                    pf_raw["dis24"].values,
                )
            )
            return cf_raw, pf_raw, expected_dis24

        return _mock_ensemble_raw

    def get_processed_data(
        self,
        country_config: CountryConfig,
        number_coord: [List[int], int] = None,
        include_step: bool = False,
        dis24: np.ndarray = None,
        single_day: bool = False,
    ) -> xr.Dataset:
        """
        Create a simplified fake processed GloFAS dataset.

        Parameters
        ----------
        country_config : CountryConfig
            Country configuration object
        number_coord : list or int, default = None
            The ensemble number coordinate
        include_step : bool, default = False
            Whether to include the forecast step coordinate
        dis24 : np.ndarray, default = None
            Optional array of discharge values, that should have the combined
            dimensions of the coordinates. If not passed, generated using
            random numbers.
        single_day: boolean, default = False
            Only create a single day of data, required for the forecasts

        Returns
        -------
        GloFAS processed xarray dataset
        """
        raw_data = self.get_raw_data(
            number_coord=number_coord,
            include_step=include_step,
            dis24=dis24,
            single_day=single_day,
        )
        coords = {}
        if number_coord is not None:
            coords = {"number": number_coord}
        if not single_day:
            coords["time"] = raw_data.time
        if include_step:
            coords["step"] = raw_data.step
        return xr.Dataset(
            {
                reporting_point.name: (
                    list(coords.keys()),
                    raw_data["dis24"]
                    .sel(
                        longitude=reporting_point.lon,
                        latitude=reporting_point.lat,
                        method="nearest",
                    )
                    .data,
                )
                for reporting_point in country_config.glofas.reporting_points
            },
            coords=coords,
        )

    @pytest.fixture()
    def mock_processed_data_reanalysis(
        self, mocker, mock_country_config
    ) -> xr.Dataset:
        """Create fake processed GloFAS reanalysis data."""
        mocker.patch(
            "aatoolbox.datasources.glofas.reanalysis.xr.load_dataset",
            return_value=self.get_raw_data(),
        )
        return self.get_processed_data(country_config=mock_country_config)

    @pytest.fixture()
    def mock_processed_data_forecast(
        self, mock_ensemble_raw, mocker, mock_country_config
    ):
        """Create fake processed GloFAS forecast or reforecast data."""

        def _mock_processed_dta_forecast(single_day: bool = False):
            cf_raw, pf_raw, expected_dis24 = mock_ensemble_raw(
                single_day=single_day
            )
            mocker.patch(
                "aatoolbox.datasources.glofas.forecast.xr.load_dataset",
                side_effect=[cf_raw, pf_raw],
            )
            return self.get_processed_data(
                country_config=mock_country_config,
                number_coord=self.numbers,
                include_step=True,
                dis24=mock_ensemble_raw(single_day=single_day)[2],
                single_day=single_day,
            )

        return _mock_processed_dta_forecast

    def test_reanalysis_process(
        self, mock_country_config, mock_processed_data_reanalysis
    ):
        """Test GloFAS reanalysis process method."""
        glofas_reanalysis = GlofasReanalysis(
            country_config=mock_country_config,
            geo_bounding_box=self.geo_bounding_box,
            start_date=date(year=2021, month=1, day=1),
            end_date=date(year=2021, month=12, day=31),
        )
        output_filepath = glofas_reanalysis.process()[0]
        # use open_dataset since load_dataset is patched
        with xr.open_dataset(output_filepath) as output_ds:
            assert output_ds.equals(mock_processed_data_reanalysis)

    def test_reforecast_process(
        self, mock_country_config, mock_processed_data_forecast
    ):
        """Test GloFAS reforecast process method."""
        target_dataset = mock_processed_data_forecast()
        glofas_reforecast = GlofasReforecast(
            country_config=mock_country_config,
            geo_bounding_box=self.geo_bounding_box,
            leadtime_max=3,
            start_date=date(year=2018, month=1, day=1),
            end_date=date(year=2018, month=1, day=31),
        )
        output_filepath = glofas_reforecast.process()[0]
        with xr.open_dataset(output_filepath) as output_ds:
            assert output_ds.equals(target_dataset)

    def test_forecast_process(
        self,
        mock_country_config,
        mock_processed_data_forecast,
        mock_ensemble_raw,
    ):
        """Test GloFAS forecast process method."""
        target_dataset = mock_processed_data_forecast(single_day=True)
        glofas_forecast = GlofasForecast(
            country_config=mock_country_config,
            geo_bounding_box=self.geo_bounding_box,
            leadtime_max=3,
            start_date=date(year=2022, month=1, day=1),
            end_date=date(year=2022, month=1, day=1),
        )
        output_filepath = glofas_forecast.process()[0]
        with xr.open_dataset(output_filepath) as output_ds:
            assert output_ds.equals(target_dataset)



"""Class to download and load CHIRPS observational precipitation data."""
import calendar
import logging
import ssl
from abc import abstractmethod
from datetime import date, datetime
from pathlib import Path
from typing import Optional, Union
from urllib.request import urlopen

import cftime
import pandas as pd
import requests
import xarray as xr

from aatoolbox.config.countryconfig import CountryConfig
from aatoolbox.datasources.datasource import DataSource
from aatoolbox.utils.check_file_existence import check_file_existence
from aatoolbox.utils.dates import get_date_from_user_input
from aatoolbox.utils.geoboundingbox import GeoBoundingBox

logger = logging.getLogger(__name__)

_FIRST_AVAILABLE_DATE = date(year=1981, month=1, day=1)

_VALID_RESOLUTIONS = (0.05, 0.25)

_BASE_URL = "https://iridl.ldeo.columbia.edu/SOURCES/.UCSB/.CHIRPS/.v2p0/"


class _Chirps(DataSource):
    """
    Base class object to retrieve CHIRPS observational precipitation data.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    geo_bounding_box: GeoBoundingBox
        the bounding coordinates of the area that should be included in the
        data.
    frequency: str
        time resolution of the data to be downloaded. It can be "daily" or
        "monthly".
    resolution: float, default = 0.05
        resolution of data to be downloaded. Can be 0.05 or 0.25.
    start_date: Optional[Union[datetime.date, str]], default = None
        Data will be considered starting from date `start_date`.
        Input can be an ISO8601 string or `datetime.date` object.
        If None, it is set to 1981-1-1.
    end_date: Optional[Union[datetime.date, str]], default = None
        Data will be considered up to date `end_date`.
        Input can be an ISO8601 string or `datetime.date` object.
        If None, it is set to the date for which most recent data
        is available.
    """

    @abstractmethod
    def __init__(
        self,
        country_config: CountryConfig,
        geo_bounding_box: GeoBoundingBox,
        frequency: str,
        date_range_freq: str,
        resolution: float = 0.05,
        start_date: Optional[Union[date, str]] = None,
        end_date: Optional[Union[date, str]] = None,
    ):
        super().__init__(
            country_config=country_config,
            datasource_base_dir="chirps",
            is_public=True,
        )

        # round coordinates to correspond with the grid the CHIRPS data is on
        # non-rounded coordinates can be given to the URL which then
        # automatically rounds them, but for file saving we prefer to do
        # this ourselves
        self._geobb = geo_bounding_box.round_coords(round_val=resolution)
        self._frequency = frequency
        self._resolution = resolution
        self._date_range_freq = date_range_freq

        if start_date is None:
            start_date = _FIRST_AVAILABLE_DATE
        if end_date is None:
            end_date = self._get_last_available_date()

        self._start_date = get_date_from_user_input(start_date)
        self._end_date = get_date_from_user_input(end_date)

        self._check_dates_validity()

        if resolution not in _VALID_RESOLUTIONS:
            raise ValueError(
                f"The given resolution is {self._resolution}, which is "
                "not available. Has to be 0.05 or 0.25."
            )

    def download(  # type: ignore
        self,
        clobber: bool = False,
    ):
        """
        Download the CHIRPS observed precipitation as NetCDF file.

        Parameters
        ----------
        clobber : bool, default = False
            If True, overwrites existing raw files

        Returns
        -------
        The folder where the data is downloaded.
        """
        # Create a list of date tuples
        date_list = self._create_date_list(logging_level=logging.INFO)

        # Data download
        for d in date_list:
            last_filepath = self._download_prep(d=d, clobber=clobber)

        return last_filepath.parents[0]

    def process(self, clobber: bool = False):
        """
        Process the CHIRPS data.

        Should only be called after data has been download.

        Parameters
        ----------
        clobber : bool, default = False
            If True, overwrites existing processed files.

        Returns
        -------
        The folder where the data is processed.
        """
        # Create a list with all raw data downloaded
        filepath_list = self._get_to_be_processed_path_list()

        for filepath in filepath_list:
            try:
                ds = xr.open_dataset(filepath, decode_times=False)
            except ValueError as err:
                raise ValueError(
                    f"The dataset {filepath} is not a valid netcdf file: "
                    "something probbly went wrong during the download. "
                    "Try downloading the file again."
                ) from err
            processed_file_path = self._get_processed_path(filepath)
            processed_file_path.parent.mkdir(parents=True, exist_ok=True)
            last_filepath = self._process(
                filepath=processed_file_path, ds=ds, clobber=clobber
            )

        return last_filepath.parents[0]

    def load(self) -> xr.Dataset:
        """
        Load the CHIRPS data.

        Should only be called after the data
        has been downloaded and processed.

        Returns
        -------
        The processed CHIRPS dataset.
        """
        # Get list of filepaths of files to be loaded

        filepath_list = self._get_to_be_loaded_path_list()

        # Merge all files in one dataset
        if not filepath_list:
            raise FileNotFoundError(
                "Cannot find any netcdf file for the chosen combination "
                "of frequency, resolution and area. Make sure "
                "sure that you have already called the 'process' method."
            )

        try:
            ds = xr.open_mfdataset(
                filepath_list,
            )
            # include the names of all files that are included in the ds
            ds.attrs["included_files"] = [f.stem for f in filepath_list]
        except FileNotFoundError as err:
            raise FileNotFoundError(
                "Cannot find one or more netcdf files corresponding "
                "to the selected range. Make sure that you already "
                "downloaded and processed those data."
            ) from err

        return ds.rio.write_crs("EPSG:4326", inplace=True)

    def _download_prep(self, d, clobber):
        # Preparatory steps for the actual download
        year = str(d.year)
        month = f"{d.month:02d}"
        day = f"{d.day:02d}"
        output_filepath = self._get_raw_path(year=year, month=month, day=day)
        output_filepath.parent.mkdir(parents=True, exist_ok=True)
        url = self._get_url(year=year, month=month, day=day)
        # Actual download
        return self._download(
            filepath=output_filepath,
            url=url,
            clobber=clobber,
        )

    def _create_date_list(self, logging_level=logging.DEBUG):
        """Create list of tuples containing the range of dates of interest."""
        date_list = pd.date_range(
            self._start_date, self._end_date, freq=self._date_range_freq
        ).tolist()

        # Create a message containing information on the downloaded data
        msg = (
            f"{self._frequency.capitalize()} "
            "data will be downloaded, starting from "
            f"{date_list[0]} to {date_list[-1]}."
        )

        logging.log(logging_level, msg)

        return date_list

    def _check_dates_validity(self):
        """Check dates vailidity."""
        end_avail_date = self._get_last_available_date()

        if (
            not _FIRST_AVAILABLE_DATE
            <= self._start_date
            <= self._end_date
            <= end_avail_date
        ):
            raise ValueError(
                "Make sure that the input dates are ordered in the following "
                f"way: {_FIRST_AVAILABLE_DATE} <= {self._start_date} <= "
                f"{self._end_date} <= {end_avail_date}. The two dates above "
                "indicate the range for which CHIRPS data are "
                "currently available."
            )

    def _get_file_name_base(
        self,
        year: str,
        month: str,
    ) -> str:
        if len(month) == 1:
            month = f"0{month}"

        file_name_base = (
            f"{self._country_config.iso3}_chirps_"
            f"{self._frequency}_{year}_{month}_"
        )

        return file_name_base

    @abstractmethod
    def _get_file_name(
        self,
        year: str,
        month: str,
        day: str,
    ) -> str:
        pass

    def _get_raw_path(self, year: str, month: str, day: str) -> Path:
        return self._raw_base_dir / self._get_file_name(
            year=year, month=month, day=day
        )

    def _get_processed_path(self, raw_path: Path) -> Path:
        return self._processed_base_dir / raw_path.parts[-1]

    def _get_location_url(self):

        location_url = (
            f"X/%28{self._geobb.lon_min}%29%28{self._geobb.lon_max}"
            f"%29RANGEEDGES/"
            f"Y/%28{self._geobb.lat_max}%29%28{self._geobb.lat_min}"
            f"%29RANGEEDGES/"
        )

        return location_url

    @abstractmethod
    def _get_url(self, year: str, month: str, day: str):
        pass

    @abstractmethod
    def _get_last_available_date(self):
        pass

    def _get_to_be_processed_path_list(self):
        """Get list of filepaths of files to be processed."""
        date_list = self._create_date_list()

        filepath_list = [
            self._get_raw_path(
                year=f"{d.year}", month=f"{d.month:02d}", day=f"{d.day:02d}"
            )
            for d in date_list
        ]
        filepath_list.sort()

        return filepath_list

    def _get_to_be_loaded_path_list(self):
        """Get list of filepaths of files to be loaded."""
        date_list = self._create_date_list()

        filepath_list = [
            self._get_processed_path(
                self._get_raw_path(
                    year=f"{d.year}",
                    month=f"{d.month:02d}",
                    day=f"{d.day:02d}",
                )
            )
            for d in date_list
        ]
        filepath_list.sort()

        return filepath_list

    @staticmethod
    def _read_csv_from_url(url):

        context = ssl.create_default_context()
        context.set_ciphers("DEFAULT")
        result = urlopen(url, context=context)

        return pd.read_csv(result)

    @check_file_existence
    def _download(self, filepath: Path, url: str, clobber: bool) -> Path:
        logger.info("Downloading CHIRPS NetCDF file.")
        response = requests.get(
            url,
        )
        with open(filepath, "wb") as out_file:
            out_file.write(response.content)
        return filepath

    @check_file_existence
    def _process(self, filepath: Path, ds, clobber: bool) -> Path:
        pass


class ChirpsMonthly(_Chirps):
    """
    Class object to retrieve CHIRPS observational monthly precipitation data.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    geo_bounding_box: GeoBoundingBox
        the bounding coordinates of the area that should be included in the
        data.
    start_date: Optional[Union[datetime.date, str]], default = None
        Data will be considered starting from date `start_date`.
        Input can be an ISO8601 string or `datetime.date` object.
        If None, it is set to 1981-1-1.
    end_date: Optional[Union[datetime.date, str]], default = None
        Data will be considered up to date `end_date`.
        Input can be an ISO8601 string or `datetime.date` object.
        If None, it is set to the date for which most recent data
        is available.

    Examples
    --------
    >>> from aatoolbox import create_country_config, CodAB, GeoBoundingBox
    >>> from aatoolbox import ChirpsMonthly
    >>> import datetime
    >>>
    >>> country_config = create_country_config(iso3="bfa")
    >>> codab = CodAB(country_config=country_config)
    >>> codab.download()
    >>> admin0 = codab.load(admin_level=0)
    >>> geo_bounding_box = GeoBoundingBox.from_shape(admin0)
    >>>
    >>> start_date = datetime.date(year=2007, month=10, day=23)
    >>> end_date = datetime.date(year=2020, month=3, day=2)
    >>> chirps_monthly = ChirpsMonthly(
    ...   country_config=country_config,
    ...   geo_bounding_box=geo_bounding_box,
    ...   start_date=start_date,
    ...   end_date=end_date
    ... )
    >>> chirps_monthly.download()
    >>> chirps_monthly.process()
    >>> chirps_monthly_data = chirps_monthly.load()
    """

    def __init__(
        self,
        country_config: CountryConfig,
        geo_bounding_box: GeoBoundingBox,
        start_date: Optional[Union[date, str]] = None,
        end_date: Optional[Union[date, str]] = None,
    ):
        super().__init__(
            country_config=country_config,
            geo_bounding_box=geo_bounding_box,
            date_range_freq="SMS",
            frequency="monthly",
            resolution=0.05,
            start_date=start_date,
            end_date=end_date,
        )

    def _get_file_name(
        self,
        year: str,
        month: str,
        day: str,
    ) -> str:

        file_name_base = self._get_file_name_base(
            year=year,
            month=month,
        )

        file_name = (
            f"{file_name_base}"
            f"r{self._resolution}"
            f"_{self._geobb.get_filename_repr(p=0)}.nc"
        )
        return file_name

    def _get_last_available_date(self):
        """Get the most recent date for which data is available."""
        # The url contains a table where the last date available is
        # specified.
        url = (
            f"{_BASE_URL}"
            ".monthly/.global/"
            ".T/last/subgrid/0./add/T/table%3A/1/%3Atable/.csv"
        )

        df = self._read_csv_from_url(url)

        datetime_object = datetime.strptime(df.values[0][0], "%b %Y")
        day = calendar.monthrange(datetime_object.year, datetime_object.month)[
            1
        ]
        datetime_object = datetime_object.replace(day=day)

        return datetime_object.date()

    def _get_url(self, year: str, month: str, day: str) -> str:

        # Convert month from month number (in string format) to
        # three-letter name
        month_name = calendar.month_abbr[int(month)]

        location_url = self._get_location_url()

        url = (
            f"{_BASE_URL}"
            ".monthly/.global/.precipitation/"
            f"{location_url}"
            f"T/%28{month_name}%20{year}%29%28{month_name}%20{year}"
            "%29RANGEEDGES/data.nc"
        )

        return url

    @check_file_existence
    def _process(self, filepath: Path, ds, clobber: bool) -> Path:
        # fix dates
        ds.aat.correct_calendar(inplace=True)
        ds = xr.decode_cf(ds)
        if "prcp" in list(ds.keys()):
            ds = ds.rename({"prcp": "precipitation"})
        xr.Dataset.to_netcdf(ds, path=filepath)
        return filepath


class ChirpsDaily(_Chirps):
    """
    Class object to retrieve CHIRPS observational monthly precipitation data.

    Parameters
    ----------
    country_config : CountryConfig
        Country configuration
    geo_bounding_box: GeoBoundingBox
        the bounding coordinates of the area that should be included in the
        data.
    resolution: float, default = 0.05
        resolution of data to be downloaded. Can be 0.05 or 0.25.
    start_date: Optional[Union[datetime.date, str]], default = None
        Data will be considered starting from date `start_date`.
        Input can be an ISO8601 string or `datetime.date` object.
        If None, it is set to 1981-1-1.
    end_date: Optional[Union[datetime.date, str]], default = None
        Data will be considered up to date `end_date`.
        Input can be an ISO8601 string or `datetime.date` object.
        If None, it is set to the date for which most recent data
        is available.

    Examples
    --------
    >>> from aatoolbox import create_country_config, CodAB, GeoBoundingBox,
    ... ChirpsDaily
    >>> import datetime
    >>>
    >>> country_config = create_country_config(iso3="bfa")
    >>> codab = CodAB(country_config=country_config)
    >>> codab.download()
    >>> admin0 = codab.load(admin_level=0)
    >>> geo_bounding_box = GeoBoundingBox.from_shape(admin0)
    >>> start_date = datetime.date(year=2007, month=10, day=23)
    >>> end_date = datetime.date(year=2020, month=3, day=2)
    >>> chirps_daily = ChirpsDaily(
    ...   country_config=country_config,
    ...   geo_bounding_box=geo_bounding_box,
    ...   start_date=start_date,
    ...   end_date=end_date
    ... )
    >>> chirps_daily.download()
    >>> chirps_daily.process()
    >>> chirps_daily_data = chirps_daily.load()
    """

    def __init__(
        self,
        country_config: CountryConfig,
        geo_bounding_box: GeoBoundingBox,
        resolution: float = 0.05,
        start_date: Optional[Union[date, str]] = None,
        end_date: Optional[Union[date, str]] = None,
    ):
        super().__init__(
            country_config=country_config,
            geo_bounding_box=geo_bounding_box,
            date_range_freq="D",
            frequency="daily",
            resolution=resolution,
            start_date=start_date,
            end_date=end_date,
        )

    def _get_file_name(
        self,
        year: str,
        month: str,
        day: str,
    ) -> str:

        file_name_base = self._get_file_name_base(
            year=year,
            month=month,
        )

        if len(day) == 1:
            day = f"0{day}"

        file_name = (
            f"{file_name_base}"
            f"{day}_"
            f"r{self._resolution}"
            f"_{self._geobb.get_filename_repr(p=0)}.nc"
        )
        return file_name

    def _get_last_available_date(self):
        """Get the most recent date for which data is available."""
        # The url contains a table where the last date available is
        # specified.
        url = (
            f"{_BASE_URL}"
            ".daily-improved/.global/"
            f"{str(self._resolution).replace('.', 'p')}/"
            ".T/last/subgrid/0./add/T/table%3A/1/%3Atable/.csv"
        )

        df = self._read_csv_from_url(url)

        datetime_object = datetime.strptime(df.values[0][0], "%d %b %Y")

        return datetime_object.date()

    def _get_url(self, year: str, month: str, day: str) -> str:

        # Convert month from month number (in string format) to
        # three-letter name
        month_name = calendar.month_abbr[int(month)]

        location_url = self._get_location_url()

        url = (
            f"{_BASE_URL}"
            ".daily-improved/.global/."
            f"{str(self._resolution).replace('.', 'p')}/.prcp/"
            f"{location_url}"
            f"T/%28{day}%20{month_name}%20{year}%29%28{day}"
            f"%20{month_name}%20{year}"
            "%29RANGEEDGES/data.nc"
        )

        return url

    @check_file_existence
    def _process(self, filepath: Path, ds, clobber: bool) -> Path:
        # fix dates
        ds = ds.assign_coords(
            T=cftime.datetime.fromordinal(
                ds.T.values, calendar="standard", has_year_zero=False
            )
        )
        ds.aat.correct_calendar(inplace=True)
        if "prcp" in list(ds.keys()):
            ds = ds.rename({"prcp": "precipitation"})
        xr.Dataset.to_netcdf(ds, path=filepath)
        return filepath
